# Declare top-level shortcuts
from mysqlorm.ORMModel import ORMModel
from mysqlorm.MySQLConnect import MySQLConnect
from mysqlorm.QueryBuilder import QueryBuilder
from mysqlorm.ResaultBuilder import ResaultBuilder


""" mysqldb connect package mysqlorm """
""" mysql resault builder            """

from mysqlorm.MySQLConnect import MySQLConnect


class ResaultBuilder(object):
    """ Resault Builder """

    @classmethod
    def execute(cls, sql, parameter=(), many=False):
        """ execute """
        return MySQLConnect.execute(sql, parameter=parameter, many=False)

    @classmethod
    def query(cls, builder, few=True):
        """ query """
        resault = MySQLConnect.execute(builder.BuildQuerySQLString(), few=few)

        if not resault:
            return () if few else None

        return tuple(map(lambda attributes: cls.buildORMModelInstance(builder._ormmodel, attributes), resault)) \
            if few else cls.buildORMModelInstance(builder._ormmodel, resault)

    @classmethod
    def buildORMModelInstance(cls, model, attributes):
        """ build model instance """
        return model(origin_attributes=attributes)



""" mysqldb connect package mysqlorm """
""" mysql connect operation library  """

import time
import pymysql as MySQLdb


class MySQLConnect(object):
    """
    MySQL connect, set db connect, example:
    db_config = {
        "host": "localhost",
        "user": "user",
        "passwd": "passwd",
        "db": "schema",
    }
    MySQLConnect.connect(db_config)
    """

    mysql_connect_count = 0
    mysql_connect_config = {}
    sql_statement_log = []
    last_executed = ""
    db_connect = {}
    db = {}

    @classmethod
    def getDB(cls):
        return cls.db

    @classmethod
    def getDBConnect(cls):
        return cls.db_connect

    @classmethod
    def connect(cls, db_config={}):
        """ connect """
        if len(db_config):
            db_config['cursorclass'] = MySQLdb.cursors.DictCursor
            cls.mysql_connect_config = db_config

        cls.db_connect = MySQLdb.connect(**cls.mysql_connect_config)
        cls.db = cls.db_connect.cursor()
        cls.mysql_connect_count += 1

    @classmethod
    def __dbexecute(cls, sql, parameter, many):
        """ db execute """
        if many:
            cls.db.executemany(sql, parameter)
        else:
            cls.db.execute(sql, parameter)

    @classmethod
    def __fetch(cls, sql, few):
        """ fetch all """
        if sql.upper().replace(" ", "").startswith("SELECT"):
            return cls.db.fetchall() if few else cls.db.fetchone()

        return cls.db_connect.commit()

    @classmethod
    def __check_connect(cls):
        """ check connect """
        try:
            cls.db_connect.ping()
        except:
            cls.connect()
            time.sleep(10)
            cls.__check_connect()

    @classmethod
    def execute(cls, sql, parameter=(), many=False, few=True):
        """ execute sql """
        cls.__check_connect()
        cls.__dbexecute(sql, parameter, many)
        cls.sql_statement_log.append({time.time(): sql})

        return cls.__fetch(sql, few)

    @classmethod
    def rollback(cls):
        """ mysql rollback """
        return cls.db_connect.rollback()


import sys
from setuptools import setup, find_packages
from os import path
from io import open


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = ['pymysql',] if sys.version_info[0] == 3 else ['MySQLdb',]

setup(
    name = 'mysqlorm',
    version = '0.1',
    description = 'A simple ORM MySQL operation Library, running on Python3.',
    long_description = long_description,
    long_description_content_type='text/markdown',
    url = 'https://github.com/cirlmcesc/mysqlorm',
    author = 'cirlmcesc',
    author_email = 'cirlmcesc_ma@163.com',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords = 'mysql orm',
    packages = find_packages(),
    install_requires = install_requires,
    extras_require={
        'dev': [],
        'test': [],
    },
    package_data={},
    data_files=[],
    entry_points={},
    project_urls = {},
)


""" mysqldb connect package mysqlorm """
""" mysql orm model                  """

from mysqlorm.MySQLConnect import MySQLConnect
from mysqlorm.QueryBuilder import QueryBuilder
from mysqlorm.ResaultBuilder import ResaultBuilder


class ORMModel(object):
    """ ORM Model """

    no_attribute = (
        "fields", "table_name", "_ORMModel__id", "origin_attributes", "origin_id")

    def __init__(self, table_name="", attributes={}, origin_attributes={}):
        if not isinstance(attributes, dict) and not isinstance(origin_attributes, list):
            raise AttributeError("data type error.")

        self.fields = []

        if self.table_name == "" and table_name == "":
            self.table_name = self.__class__.__name__.lower()
        elif table_name != "":
            self.table_name = table_name

        def setModelInstanceAttribute(attributedata):
            for val in attributedata:
                self.__setattr__(val, attributedata.get(val, None))

        if isinstance(attributes, dict) and len(attributes):
            setModelInstanceAttribute(attributes)

        if isinstance(origin_attributes, dict) and len(origin_attributes):
            self.origin_attributes = origin_attributes
            self.origin_id = origin_attributes.get("id", "")
            setModelInstanceAttribute(origin_attributes)
        else:
            self.origin_attributes = origin_attributes
            self.origin_id = 0

    def __setattr__(self, attr, value):
        if not attr in self.no_attribute and not attr in self.fields:
            fields = self.fields
            fields.append(attr)
            self.__dict__["fields"] = fields

        self.__dict__[attr] = value

    def __getattribute__(self, attr):
        new_instance = super(ORMModel, self)

        if attr == "table_name" and new_instance.__getattribute__("table_name") == "":
            return new_instance.__class__.__name__.lower()

        try:
            return new_instance.__getattribute__(attr)
        except AttributeError:
            return None

    @classmethod
    def get(cls):
        """ get all data """
        return cls.all()

    @classmethod
    def all(cls):
        """ get all data """
        return QueryBuilder(cls).get()

    @classmethod
    def first(cls):
        """ get first data """
        return ResaultBuilder.query(QueryBuilder(cls).limit(1), few=False)

    @classmethod
    def update(cls, data):
        """ update data """
        return QueryBuilder(cls).update(data)

    @classmethod
    def insert(cls, data):
        """ insert data """
        return QueryBuilder(cls).insert(data)

    @classmethod
    def find(cls, data_id):
        """ get the id data """
        return QueryBuilder(cls).where('id', data_id).first()

    @classmethod
    def select(cls, *fields):
        """ select """
        return QueryBuilder(cls).select(*fields)

    @classmethod
    def where(cls, field, *args, **kw):
        """ new QueryBuilder and .where """
        return QueryBuilder(cls).where(field, *args, **kw)

    @classmethod
    def has(cls, field, *args, **kw):
        """ check has """
        return QueryBuilder(cls).where(field, *args, **kw).has()

    @classmethod
    def query(cls, sql):
        """ execute sql string """
        return MySQLConnect.execute(sql)

    def _arrangeAttributes(self):
        """ arrange self attribute """
        attributes = dict().fromkeys(self.fields)

        for attr in attributes:
            if attr != '_ORMModel__id':
                attributes[attr] = getattr(self, attr)

        return attributes

    def save(self):
        """ update model attribute change """
        attributes = self._arrangeAttributes()

        if len(self.origin_attributes):
            QueryBuilder(self).where("id", self.origin_id).update(attributes)

            return self.find(self.origin_id)

        return self.insert(attributes)

    def delete(self):
        """ delete this model data """
        if self.origin_id == 0:
            raise AttributeError("Instance have not attribute 'ID'")

        return QueryBuilder(self).where("id", self.origin_id).delete()

    def dict(self):
        """ self to dict """
        d = {}

        for field in self.fields:
            d[field] = getattr(self, field)

        return d



""" mysqldb connect package mysqlorm  """
""" mysql query statement builder     """

import random
from inspect import isfunction
from mysqlorm.ResaultBuilder import ResaultBuilder
from mysqlorm.MySQLConnect import MySQLConnect


class QueryBuilder(object):
    """ Query Builder """

    _selectstr = "SElECT %s FROM `%s`"
    _updatestr = "UPDATE `%s` SET "
    _deletestr = "DELETE FROM `%s`"
    _insertstr = "INSERT INTO `%s` %s VALUES %s"

    def __init__(self, ormmodel):
        self._ormmodel = ormmodel
        self._table_name = ormmodel.table_name
        self._fields = "`*`"
        self._join = []
        self._where = []
        self._order = []
        self._group = []
        self._limit = ""
        self._like_value = []

    def select(self, *fields):
        """ set select fields """
        if len(fields) > 0:
            self._fields = "`id`, " + (", ".join(list(map(
                lambda f: self._TableFieldAddGrave(f) if "." in f else "`%s`" % f, fields))))

        return self

    def BuildQuerySQLString(self):
        """ build sql string """
        if isinstance(self._fields, list):
            field = tuple(self._fields).__str__().replace('(', '')
            field = field.replace(')', '')
        else:
            field = self._fields

        sql = self._selectstr % (field.replace("'", ''), self._table_name)

        return sql + "".join(map(lambda condition: self._buildConditionString(condition), [
            "_join", "_where", "_group", "_order", "_limit"
        ]))

    def _transConditionString(self, condition):
        """ transformation condition string """
        if condition == "_order":
            return " ORDER BY "
        elif condition == "_group":
            return " GROUP BY "

        return "%s " % condition.replace("_", " ").upper()

    def _buildConditionString(self, condition):
        """ build condition string """
        content = getattr(self, condition)
        joinstring = ", " if condition == "_order" or condition == "_group" else " "
        startstring =  " " if condition == "join" else self._transConditionString(condition)

        if isinstance(content, str) and len(content):
            return startstring + content

        return (startstring + joinstring.join(content)) if len(content) else ""

    def join(self, table, current_table_field, target_table_field, **kw):
        """ join """
        self._join.append("%s %s ON %s = %s" % (
            "INNER JOIN" if not len(kw) else kw["_cmp"],
            table, current_table_field, target_table_field))

        return self

    def leftJoin(self, table, current_table_field, target_table_field):
        """ left join """
        return self.join(table, current_table_field, target_table_field, _cmp="LEFT JOIN")

    def rightJogin(self, table, current_table_field, target_table_field):
        """ right join """
        return self.join(table, current_table_field, target_table_field, _cmp="RIGHT JOin")

    def _ConnectWhereString(self, whereCmp):
        if len(self._where) > 0 and self._where[-1] != "AND" and self._where[-1] != "OR":
            self._where.append(whereCmp)

    def _appendBuildWhereString(self, field, *args, **kw):
        _cmp = "=" if len(args) == 1 else args[0]
        value = args[0] if len(args) == 1 else args[1]
        restring = value.replace(" ", "") if isinstance(value, str) else str(value)
        field = self._TableFieldAddGrave(field) if "." in field else "`%s`" % field

        if (not restring.startswith("(") or
            not restring.startswith("'") or
            not restring.startswith("NULL")):
            valuestr = "%s" if isinstance(value, int) and isinstance(value, float) else "'%s'"
            value = valuestr % value

        self._where.append(" ".join((field, _cmp, value)))

        return self

    def where(self, field, *args, **kw):
        """ where """
        self._ConnectWhereString("AND" if not len(kw) else kw['_cmp'])

        if isfunction(field):
            builder = field(QueryBuilder(self._ormmodel))
            self._where.append("(%s)" % builder._buildConditionString("_where").replace(" WHERE ", ""))
        elif isinstance(field, list) or isinstance(field, tuple):
            for condition in field:
                self.where(*condition)
        else:
            if len(args) > 1 and "LIKE" in args[0].upper():
                self._whereLike(field, args[1],
                    _cmp="NOT LIKE" if "NOT" in args[0].upper() else "LIKE")
            else:
                self._appendBuildWhereString(field, *args, **kw)

        return self

    def when(self, booleanvalue, condition):
        """ execute when the Boolean value is true """
        if booleanvalue:
            return condition(self)

        return self

    def _TableFieldAddGrave(self, field):
        """ field add grave """
        field = field.split(".")
        field[len(field) - 1] = "`%s`" % field[len(field) - 1]

        return ".".join(field)

    def _whereLike(self, field, value, _cmp):
        """ like """
        field = self._TableFieldAddGrave(field) if "." in field else "`%s`" % field
        self._where.append("LOCATE('%s', %s) > 0" % (value, field))

        return self

    def orWhere(self, field, *args):
        """ or where """
        return self.where(field, *args, _cmp="OR")

    def whereIn(self, field, _tuple, **kw):
        """ where in """
        _cmp = "IN" if not len(kw) else kw["_cmp"]
        return self.where(field, _cmp, "('%s')" % ("' ,'".join(_tuple)))

    def whereNotIn(self, field, _tuple):
        """ where not in """
        return self.whereIn(field, _tuple, _cmp="NOT IN")

    def whereBetween(self, field, from_condition, to_condition, **kw):
        """ where between """
        between = "BETWEEN" if not len(kw) else kw["_cmp"]

        return self.where(field, between, "'%s' AND '%s'" % (from_condition, to_condition))

    def whereNotBetween(self, field, from_condition, to_condition):
        """ where not between """
        return self.whereBetween(field, from_condition, to_condition, _cmp="NOT BETWEEN")

    def whereNull(self, field, **kw):
        """ where field null """
        _cmp = "IS" if not len(kw) else kw["_cmp"]

        return self.where(field, _cmp, "NUll")

    def whereNotNull(self, field):
        """ where field is not null"""
        return self.whereNull(field, _cmp="IS NOT")

    def limit(self, offset, *args):
        """ limit """
        number = offset if not len(args) else args[0]
        offset = 0 if not len(args) else offset
        self._limit = "%d, %d" % (offset, number)

        return self

    def orderBy(self, field, *args):
        """ order by """
        _cmp = "ASC" if not len(args) else "DESC"
        self._order.append(" ".join((field, _cmp)))

        return self

    def groupBy(self, field):
        """ group by """
        if isinstance(field, str):
            self._group.append(field)
        elif isinstance(field, list):
            self._group = field

        return self

    def get(self):
        """ query data """
        return ResaultBuilder.query(self)

    def first(self):
        """ query first one col data """
        return ResaultBuilder.query(self, few=False)

    def has(self):
        """ check has """
        return len(ResaultBuilder.query(self)) > 0

    def update(self, attributes):
        """ execute update """
        source_fields = ["%s = %s" % (attr, "%s") for attr in attributes]
        srouce_value = [attributes.get(attr, '') for attr in attributes]
        sql = (self._updatestr % self._table_name) + ", ".join(source_fields)
        sql = sql + self._buildConditionString("_where")

        return MySQLConnect.execute(sql, parameter=srouce_value)

    def delete(self):
        """ execute delete """
        sql = self._deletestr % self._table_name
        sql = sql + self._buildConditionString("_where")

        return MySQLConnect.execute(sql)

    def insert(self, data):
        """ attribute insert to table"""
        if (not isinstance(data, list) and
            not isinstance(data, dict)) or not len(data):
            raise AttributeError("data type error.")

        fields = data.keys() if isinstance(data, dict) else random.choice(data).keys()
        sql = self._insertstr % (self._table_name, self._buildInserFieldsString(fields),
            "(%s)" % ", ".join(["%s" for index in range(len(fields))]))

        return MySQLConnect.execute(sql,
            parameter=self.__buildInsertValueParam(data),
            many=isinstance(data, list))

    def __buildInsertValueParam(self, data):
        """ build param """
        return self._builddata(data) if isinstance(data, dict) else [
            self._builddata(instance) for instance in data
        ]

    def _builddata(self, value):
        """ build param data """
        return tuple([value.get(field, '') for field in value])

    def _buildInserFieldsString(self, fields):
        """ build field string """
        fields = tuple(fields)

        if len(fields) == 1:
            return "(%s)" % fields[0]
        else:
            tuple(fields).__str__().replace("'", '')

            return fields


import os, random
import sys; sys.path.append('/Users/cirlmcescma/www/python/mysqlorm')
from mysqlorm import ORMModel, MySQLConnect


mysql_connect_config = {
    "host": "127.0.0.1",
    "user": "root",
    "passwd": "qwe123",
    "db": "ormmodel_test",
}

class TestModel(ORMModel):
    table_name = "ormmodel_test"

def test_connect():
    MySQLConnect.connect(mysql_connect_config)
    MySQLConnect.db_connect.ping()

def test_insert():
    TestModel.insert({"test": "test ormmodel single insert"})
    TestModel.insert([{"test": "test ormmodel batch insert"}, {"test": "test ormmodel batch insert"}])
    model = TestModel(attributes={"test": "test ormmodel instance save insert"})
    model.save()

def test_query():
    TestModel.get()
    TestModel.first()
    TestModel.find(random.choice(range(1, len(TestModel.get()))))
    TestModel.where('id', ">", 1).where('test', 'like', 'test').orderBy('id').limit(11).get()

def test_update():
    TestModel.update({"test": "batch update"})

    for instance in TestModel.all():
        instance.test = "instance update"
        instance.save()

def test_delete():
    for instance in TestModel.all():
        instance.delete()


if __name__ == "__main__":
    test_connect()
    test_insert()
    test_query()
    test_update()
    test_delete()
    print(MySQLConnect.sql_statement_log)

